(ns age-of-sail.core
  (:require [age-of-sail.vec2 :refer :all]
            [seesaw.core :refer :all]
            [seesaw.bind :as b]
            [clojure.math :as math]))

(def tickrate 100)
(def hardcoded-wind [0.1 3.0]) ;; A strong easternly wind!
(defonce ships (atom []))
(defonce program (atom nil))

(defn find-ship
  [ships name]
  (some #(when (= name (:name @%)) %) ships))

(defn downwind-force
  "Calculates the force for a single square downwind sail
  Force = furl × length² × dot(wind, heading)"
  [sail wind heading]
  (scale heading (* (:furl sail) (:length sail) (:length sail) (dot wind heading))))

(defn downwind-sails-force
  [ship wind]
  (loop [slots (:slots ship)
         force (zero)]
    (if-let [slot (first slots)]
      (if (= :downwind-sail (:type slot))
        (recur (rest slots) (add force (downwind-force slot wind (:heading ship))))
        (recur (rest slots) force))
      force)))

(defn physics-step
  [ship wind]
  (dosync
   ;; Update position
   (alter ship update :position add (scale (:velocity @ship) (/ tickrate)))
   ;; linear dampening
   (alter ship update :velocity scale (- 1.0 (/ 0.5 tickrate)))
   ;; wind force
   (alter ship update :velocity add (scale (downwind-sails-force @ship wind) (/ tickrate)))))

(defn tick
  [ships]
  (doseq [ship ships]
    (-> ship (physics-step hardcoded-wind))))

(defn stop-program
  []
  (when @program
    (reset! @program false)))

(defn start-program
  []
  (stop-program)
  (let [thread-continue (atom true)]
    (.start
     (Thread.
      (fn []
        (while @thread-continue
          (tick @ships)
          (Thread/sleep (quot 1000 tickrate))))))
    (reset! program thread-continue)))

;; UI
(def tracked-ship (atom nil))

(defn ship-start-stop
  []
  (letfn [(indicator [running] (if running "Stop" "Start"))
          (toggle-prgm [_] (if @@program (stop-program) (start-program)))]
    (let [btn (button :text (indicator @@program))]
      (b/bind @program (b/transform indicator) (b/property btn :text))
      (listen btn :mouse-clicked toggle-prgm)
      btn)))

(defn ship-chooser
  []
  (let [name (text :columns 20)]
    (b/bind (.getDocument name) (b/transform #(find-ship @ships %)) tracked-ship)
    (flow-panel :items ["Ship Name" name])))

(defn format-position
  [ship]
  (apply format "X: %.2f Y: %.2f" (:position @ship)))

(defn ship-info
  []
  (let [panel (vertical-panel :visible? false)
        position (label)]
    (b/bind tracked-ship (b/some identity)
            (b/tee
             (b/bind (b/transform format-position) (b/value position))))
    (add! panel (flow-panel :items ["Position" position]))
    panel))

(defn show-ui
  []
  (let [root (frame :title "Age of Sail" :content (vertical-panel))
        start-stop (ship-start-stop)
        chooser (ship-chooser)
        info (ship-info)]
    (b/bind tracked-ship (b/transform boolean) (b/tee (b/property info :visible?)
                                                      (b/b-do [_] (pack! root))))
    (doto root
      (add! start-stop)
      (add! chooser)
      (add! info)
      pack!
      show!)))

(defn -main
  [args]
  (start-program)
  (show-ui))
